//#include "DatafoxStudioIV/precompiled.h"
// Dokumentendatei
#include "document_studio.h"
#include "file/interface/studio_setup_classes.h"
#include "file/dfcom_ini_helper.h"

#ifdef DATAFOX_STUDIO_IV
#include "common/df4_crypt_internal.h"
#endif
#include "common/utils.h"

DocumentStudio::DocumentStudio( const std::string &studioVersion, const std::string &libraryVersion )
{
    // Dokumentroot
    m_root = nullptr;

    // Speicherbelegung im Flash des aktuellen Setupimage
    m_flashSetupSize = 0;

    // Flags zum Einspielen des Dokumentes
    m_flagsLoadSetup = 0;

    // Dateiname
    m_filePath = "";

    // Aktuelle Versionen in Dateiheader eintragen
    setVersions( studioVersion, libraryVersion );
    m_fileXML.init();
}

DocumentStudio::~DocumentStudio()
{
    if ( m_root )
    {
        delete m_root;
    }
}

void DocumentStudio::SetDocumentDefault( int32_t flags )
{
    // Initialisierung des Dokumentspeichers
    m_fileAES.SetDefault( flags );
}

void DocumentStudio::SetRootDefault( SetupType setupType )
{
    // Rooteintrag neu erstellen und aufsetzen
    if ( m_root )
    {
        delete m_root;
    }

    m_root = new EntryDevice( nullptr );

    m_root->setDefault( setupType );
}

int32_t DocumentStudio::SaveCheck()
{
    /*
    int32_t i = 0;
    TCHAR fname[_MAX_PATH];
    TCHAR *ext;

    CString pathName = GetPathName();
    if (pathName == "") {  // Neue Dateiangabe

    int32_t pos = 0;
    CString fileNewName;
    GetDocTemplate()->GetDocString(fileNewName, CDocTemplate::filterExt);
    fileNewName = GetTitle() + fileNewName.Tokenize(";", pos);

    QString basePath = appSettings()->pathSetups();
    if (!basePath.endsWith('\\')) basePath.append('\\');
    m_strPathName = basePath.toLatin1().data();
    m_strPathName += fileNewName;
    return 1;
    }

    _tcscpy(fname, pathName);
    ext = &fname[_tcslen(fname) - 3];
    // Dateiendung pruefen
    while(ext[i] && i < 4) {ext[i] = _totlower(ext[i]); i++;}

    int32_t res = m_fileAES.NeededVersion();
    if (res > 1 && _tcscmp(ext, _T("aes")) == 0)
    {
    CString str;
    str = tr("Da die Menge der Setupdaten die maximale Größe für Dateien der Endung (*.aes) übersteigt,\n").toLocal8Bit();
    str += tr("wird die Datei mit der Endung (*.ael) abgespeichert und weitergeführt.\n\n").toLocal8Bit();
    str += tr("Ihre Datei mit der Endung (*.aes) bleibt im letzten gespeicherten Zustand erhalten.").toLocal8Bit();

    // Hinweis zum benötigten Dateityp geben.
    MessageBox(str,  "DatafoxStudioIV", MB_OK|MB_ICONEXCLAMATION);
    // Dateiendung anpassen
    _tcscpy(ext, _T("ael"));
    SetPathName(fname);
    }

    return res;
    */
    return 1;
}

/////////////////////////////////////////////////////////////////////////////
// DocumentStudio Serialisierung
DocumentStudio::Result DocumentStudio::BuildRootFromImage( SetupType setupType )
{
    // Dokumentroot neu aufsetzen
    SetRootDefault( setupType );

    Result result;
    try
    {
        // Struktur aufbauen
        root().readDocument( 0, *this );

        // Termination of document generation if the loaded XML file has a faulty structure.
        // Only if an XML file has been loaded.
        if ( (m_currentFileType == FileType::XML) && !m_fileXML.isValid() )
        {
            result = Result( ResultCodes::FileXML_structureError, std::vector<std::string>{""} );
        }
        else
        {
            result = MkSetupImage( root(), m_flashSetupSize );
        }
    }
    catch ( ... )
    {
        // Dokumentroot neu aufsetzen
        SetRootDefault( setupType );
        result = Result( ResultCodes::FileContentCorrupted );
        ////QMessageBox::critical( qobject_cast<QWidget *>(parent()), "DatafoxStudioIV", tr( "Der Inhalt der Datei ist beschädigt und\nkann daher nicht verarbeitet werden." ) );
    }
    return result;
}

DocumentStudio::Result DocumentStudio::documentNew( SetupType setupType )
{
    // Datei neu aufsetzen
    SetDocumentDefault();

    // Dokument neu aufsetzen
    SetRootDefault( setupType );
    return MkSetupImage( root(), m_flashSetupSize );
}

DocumentStudio::Result DocumentStudio::documentOpen_AES( const std::string &filePath, bool ignoreStudioVersion,
                                                         std::function<bool( const std::string & filePath, const std::string & backupFilePath )> shouldCreateBackup,
                                                         std::function<bool( const std::string & filePath, const SoftwarebuildVersion & version, const std::string & component )> downgradeInfo )
{
    int32_t fileError = 0;

    // Version der Setupdatei prüfen
    uint8_t eep[EEP_SIZE_SU];

    FILE *fp = fopen( filePath.c_str(), "rb" );
    if ( fp == nullptr )
    {
        return Result( ResultCodes::SetupOpeningFailed );
    }

    // EEPROM - lesen
    if ( fread( eep, sizeof( uint8_t ), sizeof( eep ), fp ) != sizeof( eep ) )
    {
        fclose( fp );
        return Result( ResultCodes::FileReadingError );
    }

    fclose( fp );

    // Alte Setupdateien unter DatafoxStudioIV - Version 04.01.07.xx werden nicht geöffnet.
    //############################
    // 2017-10-05 (BOt)
    // Ab der Version 04.03.11.00 wird eine Verschlüsselung aller Passwörter der Setupdaten eingeführt. Es werden hierbei
    // auch "Altlasten" zur Konvertierung von Werten entfernt.

    // Alte Setupdateien unter DatafoxStudioIV - Version 04.01.01.xx werden nicht geöffnet.
    //############################
    // 2011-08-23 (BOt)
    // Ab der Version 04.02.02.08 wird mit einem Dateiheader gearbeitet. In diesem liegen alle Versionsinformationen.
    // Die Version im EEP wird nur noch fuer die 4er Reihe beibehalten.
    StructFlashHeader &flash_header = *reinterpret_cast<StructFlashHeader *>(eep);
    auto studioVersion = m_fileAES.endian_reverse( flash_header.studioVersion );

    SoftwarebuildVersion version( studioVersion );
    if ( ((studioVersion & 0xFF000000) == 0xFF || version < "04.01.07.00") &&
        // Workaround fuer Falsches Ablegen der Version in 04.02.01.11 als 02.11.00.00 und 04.02.02.00-07 als 02.00-07.00.00. Muster 02.xx.00.00
         !((studioVersion & 0xFF00FFFF) == 0x02000000) )
    {
        return Result( ResultCodes::OpenUnsupportedFileVersion, std::vector<std::string>{filePath, "04.01.07.00", "04.01.10"} );
    }
    else if ( version < "04.02.02.08" )
    {
        std::string backupPath = (boost::format( "%1%/%2%.bak" ) % Utils::File::dir( filePath ) % Utils::File::name( filePath )).str();

        if ( (shouldCreateBackup != nullptr) && shouldCreateBackup( filePath, backupPath ) )
        {
            if ( !Utils::File::copy( filePath, backupPath ) )
            {
                return Result( ResultCodes::BackupCreationFailed, std::vector<std::string>{backupPath} );
            }
        }
    }

    Result result;

    try
    {
        // Datei neu aufsetzen
        SetDocumentDefault();

        bool res = true;
        int32_t flags = 0;
        if ( ignoreStudioVersion )
        {
            flags |= (FileAES::fiv_min_studio | FileAES::fiv_should_studio);
        }

    reload_label:
        // Reset result information because of fileError '-6'. There will be no check for the recommended Studio version.
        result = Result();
        // Setupdatei laden
        if ( !m_fileAES.load( filePath, &fileError, static_cast<FileAES::Enum_flag_ignore_version>(flags) ) )
        {
            switch ( fileError )
            {
                case -1:
                case -2:
                {
                    result = Result( ResultCodes::FileAES_openFailed, std::vector<std::string>{filePath} );
                    break;
                }

                case -3:
                {
                    result = Result( ResultCodes::FileAES_readFailed, std::vector<std::string>{filePath} );
                    break;
                }

                case -4:
                {
                    result = Result( ResultCodes::FileAES_checksumError, std::vector<std::string>{filePath} );
                    break;
                }

                case -5:
                {
                    SoftwarebuildVersion minVersion( m_fileAES.m_file_header.minStudioVersion );
                    result = Result( ResultCodes::FileAES_versionTooNew, std::vector<std::string>{filePath,
                                     boost::str( boost::format( "%1$02d" ) % ((minVersion >> 24) & 0xFF) ),
                                     boost::str( boost::format( "%1$02d" ) % ((minVersion >> 16) & 0xFF) ),
                                     boost::str( boost::format( "%1$02d" ) % ((minVersion >> 8) & 0xFF) ),
                                     boost::str( boost::format( "%1$02d" ) % (minVersion & 0xFF) )} );
                    break;
                }

                case -6:
                {
                    SoftwarebuildVersion shouldStudioVersion( m_fileAES.m_file_header.shouldStudioVersion );
                    if ( (downgradeInfo != nullptr) && downgradeInfo( filePath, shouldStudioVersion, "DatafoxStudioIV" ) )
                    {
                        flags |= static_cast<int32_t>(FileAES::fiv_should_studio); // ignoriere die Sollteversion des DatafoxStudioIV beim Laden.
                        goto reload_label;
                    }
                    result = Result( ResultCodes::FileAES_recommendedVersionError );
                    break;
                }

                case -7:
                {
                    SoftwarebuildVersion minVersion( m_fileAES.m_file_header.minLibraryVersion );
                    result = Result( ResultCodes::FileAES_DFComDllToOld, std::vector<std::string>{filePath,
                                     boost::str( boost::format( "%1$02d" ) % ((minVersion >> 24) & 0xFF) ),
                                     boost::str( boost::format( "%1$02d" ) % ((minVersion >> 16) & 0xFF) ),
                                     boost::str( boost::format( "%1$02d" ) % ((minVersion >> 8) & 0xFF) ),
                                     boost::str( boost::format( "%1$02d" ) % (minVersion & 0xFF) )} );
                    break;
                }

                case -8:
                {
                    SoftwarebuildVersion minVersion( m_fileAES.m_file_header.shouldLibraryVersion );
                    if ( (downgradeInfo != nullptr) && downgradeInfo( filePath, minVersion, "DFComDLL" ) )
                    {
                        flags |= static_cast<int32_t>(FileAES::fiv_should_studio); // ignoriere die Sollteversion des DatafoxStudioIV beim Laden.
                        goto reload_label;
                    }
                    result = Result( ResultCodes::FileAES_recommendedVersionError );
                    break;
                }

                case -9:
                {
                    result = Result( ResultCodes::FileAES_wrongFormat, std::vector<std::string>{filePath} );
                    break;
                }

                case -10:
                case -11:
                case -12:
                case -13:
                case -14:
                case -15:
                case -16:
                case -17:
                case -18:
                {
                    result = Result( ResultCodes::FileAES_contentChecksumError, std::vector<std::string>{filePath} );
                    break;
                }

                default:
                {
                    result = Result( ResultCodes::FileAES_internalError, std::vector<std::string>{filePath} );
                    break;
                }

                case 1: // Laden OK
                {
                    break;
                }
            }
        }

        // Weitere Verarbeitung abbrechen
        if ( !result )
        {
            SetDocumentDefault();
            return result;
        }

        // Studio-Version aus dem Flash-Header auslesen. Sollte die Version nicht gesetzt sein (z.B. Datei mit einer älteren
        // Version des Datafox Studios erstellt), dann wird die minimale Studioversion aus dem File-Header verwendet.
        uint32_t studioVersion = std::max<>( m_fileAES.endian_reverse( m_fileAES.m_flash_header.studioVersion ), m_fileAES.m_file_header.studioVersion );
        uint32_t libraryVersion = m_fileAES.m_flash_header.libraryVersion;
        if ( studioVersion == 0 )
        {
            studioVersion = m_fileAES.m_file_header.minStudioVersion;
            m_fileAES.m_flash_header.studioVersion = m_fileAES.endian_reverse(studioVersion);
        }
        
        m_fileAES.SetStudioVersion( versionFormat( studioVersion ).c_str() );
        m_fileAES.SetLibraryVersion( versionFormat( libraryVersion ).c_str() );

        // Dokument mit gelesenen Daten aufsetzen
        result = BuildRootFromImage( m_fileAES.m_flash_header.setupType );
    }
    catch ( int32_t & e )
    {
        if ( e == 815 )
        {
            //QMessageBox::critical( qobject_cast<QWidget *>(parent()), "DatafoxStudioIV",
            //                       tr( "Die Datei konnte nicht geladen werden.\n\nEs handelt sich bei der Datei entweder um keine DatafoxStudioIV - Datei\noder sie ist zu aktuell für diese DatafoxStudioIV - Version." ) );
            return Result( ResultCodes::FileLoadingError );
        }
    }
    catch ( ... )
    {
        //QMessageBox::critical( qobject_cast<QWidget *>(parent()), "DatafoxStudioIV",
        //                       tr( "Die Datei konnte nicht geladen werden.\n\nEs handelt sich bei der Datei entweder um keine DatafoxStudioIV - Datei\noder sie ist zu aktuell für diese DatafoxStudioIV - Version." ) );
        return Result( ResultCodes::FileLoadingError );
    }

    // Wurden Daten beim laden modifiziert/angepasst?
    if ( root().isTreeModified() )
    {
        root().treeClearModified();
    }

    return result;
}

DocumentStudio::Result DocumentStudio::documentOpen_XML( const std::string &filePath, bool ignoreStudioVersion,
                                                         std::function<bool( const std::string & filePath, const std::string & backupFilePath )> shouldCreateBackup,
                                                         std::function<bool( const std::string & filePath, const SoftwarebuildVersion & version, const std::string & component )> downgradeInfo )
{
    Result result( ResultCodes::Success );

    // Detect document path from ini-file (XML-section).
    std::string ini_file_path = DFComIniHelper::findIniFile();
    if ( ini_file_path.size() > 0 )
    {
        FileIni iniFile;
        char path[512]{};
        iniFile.GetString( "XML", "docPath", path, ini_file_path.c_str() );
        bool createHelpFile = (iniFile.GetUInt( "XML", "createHelpFile", 0, ini_file_path.c_str() ) == 1);
        if ( path != nullptr )
        {
            m_fileXML.configureAdditionalHelpFile( createHelpFile, path );
        }
    }

    // Analyze the contents of the XML file to build the document.
    auto xmlDoc = m_fileXML.xmlDoc();
    if ( !xmlDoc || (xmlDoc->LoadFile( filePath.c_str() ) != tinyxml2::XML_SUCCESS) )
    {
        return Result( ResultCodes::FileLoadingError );
    }

    // Datei neu aufsetzen
    SetDocumentDefault();

    // Load succesful, continue with processing.
    auto rootNode = xmlDoc->FirstChildElement( m_fileXML.nodeNameByType( FileXML::ContentType::MainNode ).c_str() );    ///< main node of xml: 'documentStudio'
    m_fileXML.setParentNode( rootNode );
    if ( rootNode == nullptr )
    {
        return Result( ResultCodes::FileContentCorrupted );
    }

    // get header information
    if ( !(m_fileXML.readContent( FileXML::ContentType::FlashHeader, m_fileAES.m_flash_header ) &&
            m_fileXML.readContent( FileXML::ContentType::FileHeader, m_fileAES.m_file_header )) )
    {
        result = Result( ResultCodes::FileReadingError );
    }

    // Adaptation of the version information due to the different byte order.
    if ( result )
    {
        // Update the Studio version information (AES-file) with the information from the xml configuration file.
        // Heads up! The order of the following two instructions is decisive.
        m_fileAES.m_studioVersion = std::max<>( m_fileAES.m_flash_header.studioVersion, m_fileAES.m_file_header.studioVersion );
        m_fileAES.m_flash_header.studioVersion = m_fileAES.endian_reverse( m_fileAES.m_studioVersion );

        // In some cases (processing with the communication library) the studio version is not written correctly to the XML.
        // For these situations it is synchronized here with the information from the AES file.
        if ( m_fileXML.studioVersion() == SoftwarebuildVersion() )
        {
            // Update the Studio version information (XML-file) with the information from the xml configuration file.
            m_fileXML.setVersions( versionFormat( m_fileAES.m_studioVersion ), versionFormat( m_fileAES.m_flash_header.libraryVersion ) );
        }
    }

    // Dokument neu aufsetzen
    if ( result )
    {
        result = Result( BuildRootFromImage( m_fileAES.m_flash_header.setupType ).code, std::vector<std::string>{filePath} );
    }
    return result;
}

DocumentStudio::Result DocumentStudio::documentOpen( const std::string &filePath, bool ignoreStudioVersion,
                                                     std::function<bool( const std::string & filePath, const std::string & backupFilePath )> shouldCreateBackup,
                                                     std::function<bool( const std::string & filePath, const SoftwarebuildVersion & version, const std::string & component )> downgradeInfo )
{
    std::string extension = boost::algorithm::to_lower_copy( Utils::File::extension( filePath ) );

    // Depending on the file format, the document is initialized.
    FileType fileType = FileType::Undefined;
    if ( extension == "aes" || extension == "ael" )
    {
        fileType = FileType::AES_AEL;
    }
    else if ( extension == "xml" )
    {
        fileType = FileType::XML;
    }

    // Set the file type of currently loaded document.
    m_currentFileType = fileType;

    Result result( ResultCodes::OpenUnsupportedFileType );
    switch ( fileType )
    {
        case FileType::AES_AEL:
        {
            result = documentOpen_AES( filePath, ignoreStudioVersion, shouldCreateBackup, downgradeInfo );
            break;
        }
        case FileType::XML:
        {
            result = documentOpen_XML( filePath, ignoreStudioVersion, shouldCreateBackup, downgradeInfo );
            break;
        }
    }

    if ( result )
    {
        // Dateiname halten.
        m_filePath = filePath;
    }
    return result;
}

bool DocumentStudio::documentSave()
{
    return documentSaveAs( m_currentFileType, filePath() );
}

bool DocumentStudio::documentSaveAs( FileType fileType, const std::string &filePath )
{
    int32_t fileError = 0;

    bool result = false;
    std::string extension = Utils::File::extension( filePath );
    boost::to_lower( extension );

    switch ( fileType )
    {
        case FileType::AES_AEL:
        {
            // Dateidaten schreiben.
            if ( extension == "aes" || extension == "ael" )
            {
                result = m_fileAES.save( filePath, &fileError );
            }
            break;
        }
        case FileType::XML:
        {
            if ( extension == "xml" )
            {
                // Adaptation of the documentation path only if the current document is converted to the XML variant.
                if ( m_currentFileType != FileType::XML )
                {
                    // Detect document path from ini-file (XML-section).
                    std::string ini_file_path = DFComIniHelper::findIniFile();
                    if ( ini_file_path.size() > 0 )
                    {
                        FileIni iniFile;
                        char path[512]{};
                        iniFile.GetString( "XML", "docPath", path, ini_file_path.c_str() );
                        bool createHelpFile = (iniFile.GetUInt( "XML", "createHelpFile", 0, ini_file_path.c_str() ) == 1);
                        if ( path != nullptr )
                        {
                            m_fileXML.configureAdditionalHelpFile( createHelpFile, path );
                        }
                    }
                }

                m_fileXML.writeFileNames( m_fileAES.m_listFileName, m_fileAES.m_listEntrance2FileName );
                result = (m_fileXML.save( filePath ) == FileXML::ErrorCodes::Success);
            }
            break;
        }
    }

    if ( result )
    {
        // Dateiname halten.
        m_filePath = filePath;
        m_currentFileType = fileType;
    }
    return result;
}

DocumentStudio::Result DocumentStudio::MkSetupImage( EntryDevice &root, uint32_t &setupSize )
{
    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (m_currentFileType != FileType::XML);

    // Dokumentspeicher initialisieren, Listenangaben jedoch beibehalten.
    SetDocumentDefault( 0x01 );

    // Pointer im EEPROM initialisieren
    m_fileAES.m_flash_header.config = m_fileAES.endian_reverse( static_cast<uint32_t>(FLASH_KONFIG_START) );
    m_fileAES.m_flash_header.list[0] = m_fileAES.endian_reverse( static_cast<uint32_t>(FLASH_LIST_START) );

    // Die Version des Setup.
    // Wird seit Studio 04.02.02.08 fix eingetragen. Nicht relevant.
    m_fileAES.m_flash_header.version = 0x0204;

    //########################################################
    // Setup - Version mit ablegen (Ab: 03.01.30)
    m_fileAES.m_flash_header.studioVersion = m_fileAES.endian_reverse( m_fileAES.m_studioVersion );

    // In dieser Version wird bei folgenden Punkten die tatsächliche Version 04.02 eingetragen.
    // Abhängig von dem Setuptypen ist eine Mindestversion des Studio's und der Bibliothek notwendig.

    // Je nach Geraet ist ertmal eine gewisse Studioversion notwendig.
    switch ( root.setupType() )
    {
        case SetupType::Exklusiv_LineIV:
        case SetupType::Flex_MasterIV:
        case SetupType::Timeboy_MasterIV:
        case SetupType::Timeboy_Mobil_PZE:
        {
            m_fileAES.NeedStudioVersion( "04.02.01.08" );
            break;
        }

        case SetupType::EVO_2_8_Pure: // EVO 2.8 Pure / EVO 3.5 Pure
        case SetupType::EVO_4_3:      // EVO 4.3 / EVO 4.6 FlexKey (ab 04.03.14)
        case SetupType::IPC_Extension: // alle Embedded-Baugruppen
        {
            m_fileAES.NeedStudioVersion( "04.03.00.08" );
            m_fileAES.NeedLibraryVersion( "04.03.00.08" );
            break;
        }

        case SetupType::IO_Box:
        case SetupType::KYO_Inloc:
        {
            m_fileAES.NeedStudioVersion( "04.03.03.01" );
            m_fileAES.NeedLibraryVersion( "04.03.03.01" );
            break;
        }

        case SetupType::Mobil_Box:
        case SetupType::Timeboy_TBDock:
        case SetupType::Mobil_TBDock:
        {
            m_fileAES.NeedStudioVersion( "04.03.06.03" );
            m_fileAES.NeedLibraryVersion( "04.03.06.03" );
            break;
        }

        case SetupType::KYO_Oneloc:
        {
            m_fileAES.NeedStudioVersion( "04.03.07.00" );
            m_fileAES.NeedLibraryVersion( "04.03.07.00" );
            break;
        }

        case SetupType::EVO_3_5_Universal:
        {
            m_fileAES.NeedStudioVersion( "04.03.11.00" );
            m_fileAES.NeedLibraryVersion( "04.03.11.00" );
            m_fileAES.UnsupportFirmware1Version();
            m_fileAES.UnsupportFirmware2Version();
            m_fileAES.NeedFirmware3Version( "04.03.11" );
            break;
        }

        case SetupType::KYO_Cenloc:
        {
            m_fileAES.NeedStudioVersion( "04.03.12.03" );
            m_fileAES.NeedLibraryVersion( "04.03.12.03" );
            m_fileAES.UnsupportFirmware1Version();
            m_fileAES.UnsupportFirmware2Version();
            m_fileAES.NeedFirmware3Version( "04.03.12" );
            break;
        }
    }

    // Energieverwaltung wurde ueberarbeitet und benoetigt eine entspechende Firmwareversion
    // Hierdurch ist das Einspielen in eine 04.02.00.yy verhindert.
    if ( (root.m_extended.powerActiv & PAF_ENABLED) > 0 )
    {
        m_fileAES.NeedFirmware2Version( "04.02.01.18" );
    }

    // Geraetetyp ablegen (Ab: 03.01.31)
    m_fileAES.m_flash_header.setupType = root.setupType();

    // Image erzeugen
    setupSize = root.writeDocument( 0, *this );

    // Adding file header information of AES-file.
    StructFileStudioHeader tmpFileHeader = m_fileAES.m_file_header;
    if ( useAESFile )
    {
        tmpFileHeader.studioVersion = m_fileXML.studioVersion();
    }
    m_fileXML.writeContent( FileXML::ContentType::FileHeader, tmpFileHeader );

    // Adding flash header information of AES-file.
    // Storage of the data in XML in normal byte order.Therefore, the byte order of the studio version
    // information must be adapted.
    m_fileAES.m_flash_header.studioVersion = m_fileAES.endian_reverse( m_fileAES.m_flash_header.studioVersion );
    StructFlashHeader tmpFlashHeader = m_fileAES.m_flash_header;
    if ( useAESFile )
    {
        tmpFlashHeader.studioVersion = m_fileXML.studioVersion();
        tmpFlashHeader.libraryVersion = m_fileXML.libraryVersion();
    }
    m_fileXML.writeContent( FileXML::ContentType::FlashHeader, tmpFlashHeader );
    m_fileAES.m_flash_header.studioVersion = m_fileAES.endian_reverse( m_fileAES.m_flash_header.studioVersion );

    Result result;
    if ( (EEP_SIZE_SU + sizeof( StructConfig ) + 0x1C + setupSize + sizeof( StructFileStudioHeader )) > (FLASH_SIZE_SU + FLASH_SIZE_SU2) )
    {
        result = Result( ResultCodes::TooMuchDataError );
    }

    return result;
}

int32_t DocumentStudio::IsAnalogValueAvailable()
{
    int32_t analogInCounter = 0;
    uint16_t bit;
    EntryIOModule &entryIOModule = rootIOModule();

    auto count = rootSignalProcessing().rootAnalogInputs().analogInputCount();
    for ( auto i = 0; i < count; ++i )
    {
        auto analogIn = rootSignalProcessing().rootAnalogInputs().analogInput( i );
        if ( analogIn.useAs )
        {
            analogInCounter++;
        }
    }

    // IOModul Aktiviert
    if ( root().m_config.ioModuleActive )
    {
        // Alle aktivierten Kanaele zaehlen.
        // Spannungskanaele
        for ( bit = 0x01; bit < 0x100; bit <<= 1 )
        {
            if ( entryIOModule.m_ioModule.voltageInputsActive & bit )
            {
                analogInCounter++;
            }
        }

        // Stromkanaele
        for ( bit = 0x01; bit < 0x100; bit <<= 1 )
        {
            if ( entryIOModule.m_ioModule.currentInputsActive & bit )
            {
                analogInCounter++;
            }
        }
    }

    return analogInCounter;
}

int32_t DocumentStudio::IsAnalogStatusAvailable()
{
    int32_t analogInCounter = 0;
    auto count = rootSignalProcessing().rootAnalogInputs().analogInputCount();
    for ( auto i = 0; i < count; ++i )
    {
        auto analogIn = rootSignalProcessing().rootAnalogInputs().analogInput( i );
        if ( analogIn.useAs == 2 )
        {
            analogInCounter++;
        }
    }

    return analogInCounter;
}

// Pruefen ob Funktion in Kombination moeglich
int32_t DocumentStudio::IsFunctionAvailable()
{
    // Alle aktivierten Zaehler auffuehren.
    int32_t iCountIndex1 = 2;
    int32_t iCountIndex2 = 3;
    int32_t iCountIndex3 = 4;

    StructDigitalInput digitalE1 = rootSignalProcessing().rootDigitalInputs().digitalInput( 0 );
    StructDigitalInput digitalE2 = rootSignalProcessing().rootDigitalInputs().digitalInput( 1 );
    StructDigitalInput digitalE3 = rootSignalProcessing().rootDigitalInputs().digitalInput( 2 );
    StructDigitalInput digitalE4 = rootSignalProcessing().rootDigitalInputs().digitalInput( 3 );

    // E1 + E3; beide als Zaehler definiert
    if ( (digitalE1.useAs == iCountIndex1 || digitalE1.useAs == iCountIndex2 || digitalE1.useAs == iCountIndex3) &&
        (digitalE3.useAs == iCountIndex1 || digitalE3.useAs == iCountIndex2 || digitalE3.useAs == iCountIndex3) )
    {
        return TRUE;
    }

    // E2 + E4; beide als Zaehler definiert
    if ( (digitalE2.useAs == iCountIndex1 || digitalE2.useAs == iCountIndex3) &&
        (digitalE4.useAs == iCountIndex1 || digitalE4.useAs == iCountIndex3) )
    {
        return TRUE;
    }

    return FALSE;
}

int32_t DocumentStudio::IsCounterAvailable()
{
    int32_t i;
    int32_t count;
    int32_t digitalInCounter = 0;
    EntryIOModule &entryIOModule = rootIOModule();
    StructDigitalInput digitalIn;

    // Alle aktivierten Zaehler auffuehren.
    int32_t iCountIndex1 = 2, iCountIndex2 = 3, iCountIndex3 = 4, iCountIndex4 = 5;
    count = rootSignalProcessing().rootDigitalInputs().digitalInputCount();
    for ( i = 0; i < count; ++i )
    {
        digitalIn = rootSignalProcessing().rootDigitalInputs().digitalInput( i );
        if ( i % 2 )
        {
            if ( digitalIn.useAs == iCountIndex1
                 || digitalIn.useAs == iCountIndex3 )
            {
                digitalInCounter++;
            }
        }
        else
        {
            if ( digitalIn.useAs == iCountIndex1
                 || digitalIn.useAs == iCountIndex2
                 || digitalIn.useAs == iCountIndex3
                 || digitalIn.useAs == iCountIndex4 )
            {
                digitalInCounter++;
            }
        }
    }

    if ( root().m_config.ioModuleActive )
    {
        for ( i = 0; i < 8; ++i )
        {
            if ( i % 2 )
            {
                if ( entryIOModule.m_digitalInput[i].useAs == iCountIndex1
                     || entryIOModule.m_digitalInput[i].useAs == iCountIndex3 )
                {
                    digitalInCounter++;
                }
            }
            else
            {
                if ( entryIOModule.m_digitalInput[i].useAs == iCountIndex1
                     || entryIOModule.m_digitalInput[i].useAs == iCountIndex2
                     || entryIOModule.m_digitalInput[i].useAs == iCountIndex3
                     || entryIOModule.m_digitalInput[i].useAs == iCountIndex4 )
                {
                    digitalInCounter++;
                }
            }
        }
    }

    return digitalInCounter;
}

int32_t DocumentStudio::IsTimerInputChainAvailable()
{
    int32_t i;
    int32_t count;
    int32_t timerCounter = 0;

    count = rootSignalProcessing().rootTimers().timerCount();
    for ( i = 0; i < count; ++i )
    {
        StructTimer timer = rootSignalProcessing().rootTimers().timer( i );
        // Steuerung durch Eingabekette?
        if ( timer.useAs == 1 )
        {
            timerCounter++;
        }
    }

    return timerCounter;
}

int32_t DocumentStudio::IsSystemMessageAvailable()
{
    int32_t i;
    int32_t count;
    int32_t counter = 0;

    count = rootSignalProcessing().rootSystemMessages().systemMessageCount();
    for ( i = 0; i < count; ++i )
    {
        StructSystemMessage systemMessage = rootSignalProcessing().rootSystemMessages().systemMessage( i );
        // Steuerung durch Eingabekette?
        if ( systemMessage.invokeInputChain )
        {
            counter++;
        }
    }

    return counter;
}

// ########################################################
// Datensatz- und Listenbeschreibungen

uint8_t DocumentStudio::GetRecordTableCount()
{
    // Anzahl RecordTable zurueckgeben
    return rootTableRecord().size();
}

EntryTable *DocumentStudio::GetRecordTable( uint8_t num )
{
    if ( num >= rootTableRecord().size() )
    {
        return nullptr;
    }

    return (EntryTable *)rootTableRecord().value( num );
}

EntryTableField *DocumentStudio::GetRecordTableField( uint8_t ucTableNum, uint8_t ucFieldNum )
{
    // Datensatzbeschreibung ermitteln
    EntryTable *pTable = GetRecordTable( ucTableNum );

    if ( pTable == nullptr || (ucFieldNum >= pTable->size()) )
    {
        return nullptr;
    }

    return (EntryTableField *)pTable->value( ucFieldNum );
}

uint8_t DocumentStudio::GetListTableCount()
{
    // Anzahl der LSB zurueckgeben
    return rootTableList().size();
}

EntryTable *DocumentStudio::GetListTable( uint8_t num )
{
    if ( num >= rootTableList().size() )
    {
        return nullptr;
    }

    return (EntryTable *)rootTableList().value( num );
}

EntryTableField *DocumentStudio::GetListTableField( uint8_t ucTableNum, uint8_t ucFieldNum )
{
    // Listenbeschreibung ermitteln
    EntryTable *pTable = GetListTable( ucTableNum );

    if ( pTable == nullptr || (ucFieldNum >= pTable->size()) )
    {
        return nullptr;
    }

    return (EntryTableField *)pTable->value( ucFieldNum );
}

uint8_t DocumentStudio::GetAccessControlTableCount()
{
    // Anzahl der Tabellen unter Zutrittskontrolle zurueckgeben
    return 8;//rootAccessControl()->size();
}

EntryTable *DocumentStudio::GetAccessControlTable( uint8_t num )
{
    if ( num >= (rootAccessControl().size() - 1) ) // Abzüglich der Eingabekette
    {
        return nullptr;
    }

    return (EntryTable *)rootAccessControl().value( num );
}

EntryTableField *DocumentStudio::GetAccessControlTableField( uint8_t ucTableNum, uint8_t ucFieldNum )
{
    // Listenbeschreibung ermitteln
    EntryTable *pTable = GetAccessControlTable( ucTableNum );

    if ( pTable == nullptr || (ucFieldNum >= pTable->size()) )
    {
        return nullptr;
    }

    return (EntryTableField *)pTable->value( ucFieldNum );
}


// ########################################################
// Helper

void DocumentStudio::GetListName( uint8_t ucNum, char *szBuf )
{
    // Name der Listenbeschreibung ermitteln
    EntryTable *pListTable = GetListTable( ucNum );

    if ( pListTable == nullptr )
    {
        return;
    }

    strcpy( szBuf, (char *)pListTable->m_table.name );
}

std::string DocumentStudio::listName( int32_t num )
{
    // Name der Listenbeschreibung ermitteln
    EntryTable *entryTable = GetListTable( num );

    if ( entryTable == nullptr )
    {
        return "";
    }

    return entryTable->m_table.name;
}

void DocumentStudio::GetAccessControlListName( uint8_t ucNum, char *szBuf )
{
    EntryTable *entryTable = GetAccessControlTable( ucNum );

    if ( entryTable == nullptr )
    {
        return;
    }

    strcpy( szBuf, (char *)entryTable->m_table.name );
}

std::string DocumentStudio::listNameAccessControl( int32_t num )
{
    EntryTable *entryTable = GetAccessControlTable( num );

    if ( entryTable == nullptr )
    {
        return "";
    }

    return entryTable->m_table.name;
}

// Gibt Position der definierten Variablen aus Array zurueck
// z.B. Definierte Variable 3 ist an Array - Position 5
int32_t DocumentStudio::globalVariableIndex( uint8_t ucDefNum )
{
    uint8_t i, found = 0;

    for ( i = 0; i < 8; i++ )
    {
        if ( root().m_globalVar.names[i][0] != 0 )
        {
            found++;
            if ( found == ucDefNum )
            {
                return i;
            }
        }
    }

    // nicht gefunden ist erste moegliche
    return 0;
}

// Gibt den Namen der Variablen zurueck
std::string DocumentStudio::globalVariableName( uint8_t num )
{
    if ( num > 7 )
    {
        return "";
    }

    return root().m_globalVar.names[num];
}

// Gibt Anzahl der definierten Variablen zurueck
uint8_t DocumentStudio::globalVariableCount()
{
    uint8_t i, count = 0;

    for ( i = 0; i < 8; i++ )
    {
        if ( root().m_globalVar.names[i][0] != 0 )
        {
            count++;
        }
    }

    return count;
}

// ########################################################
// Ermitteln von Informationen zu Tabellen

// Ermitteln der Anzahl definierter Tabellen
int32_t DocumentStudio::GetTableCount( uint8_t typeTable )
{
    // Je nach Typ.
    switch ( typeTable )
    {
        case TT_RECORD:
            return GetRecordTableCount();

        case TT_LIST:
            return GetListTableCount();

        case TT_ACCESS_CONTROL:
            return GetAccessControlTableCount();
    }

    // Typ unbekannt.
    return 0;
}

// Ermitteln der Informationen einer Tabelle
int32_t DocumentStudio::GetTable( uint8_t num, StructTable &table, uint8_t typeTable )
{
    EntryTable *pTable = nullptr;

    // Je nach Typ.
    switch ( typeTable )
    {
        case TT_RECORD:
            pTable = GetRecordTable( num );
            break;

        case TT_LIST:
            pTable = GetListTable( num );
            break;

        case TT_ACCESS_CONTROL:
            pTable = GetAccessControlTable( num );
            break;
    }

    if ( pTable != nullptr )
    {
        table.number = num;
        // Anzahl Felder uebergeben
        table.count = pTable->m_table.count;
        // Name uebergeben
        strcpy( table.name, (const char *)pTable->m_table.name );
    }

    return (pTable != nullptr);
}

// Ermitteln der Informationen eines Tabellenfeldes
int32_t DocumentStudio::GetTableField( uint8_t numTable, uint8_t numField,
                                       StructTableField &field, uint8_t typeTable )
{
    EntryTableField *pTableField = nullptr;

    // Je nach Typ.
    switch ( typeTable )
    {
        case TT_RECORD:
            pTableField = GetRecordTableField( numTable, numField );
            break;

        case TT_LIST:
            pTableField = GetListTableField( numTable, numField );
            break;

        case TT_ACCESS_CONTROL:
            pTableField = GetAccessControlTableField( numTable, numField );
            break;
    }

    if ( pTableField != nullptr )
    {
        // Name uebergeben
        strcpy( field.name, (const char *)pTableField->m_tableField.name );
        // Typ uebergeben
        field.type = pTableField->m_tableField.type;
        // Laenge uebergeben
        field.size = pTableField->m_tableField.size;
    }

    return (pTableField != nullptr);
}

/////////////////////////////////////////////////////////////////////////////
// DocumentStudio Befehle

uint32_t DocumentStudio::GetFlashSetupSize()
{
    // ##############################################################
    // Pointer sind direkte Adressen die im AEIII+:
    //   1. Im Motorolla - Format vorliegen.
    //   2. Mit einem Basisoffset -FLASH_KONFIG_START- berechnet sind.

    // Pointer auf ListenBeschreibungen ermitteln
    uint32_t pLB = m_fileAES.endian_reverse( m_fileAES.m_flash_header.listStruct );
    pLB -= FLASH_KONFIG_START;

    // Pointerarray der Listenbeschreibungen durchschreiten
    int32_t i;
    for ( i = 0; i < MAX_RECORD_DESCRIPTION; i++ )
    {
        // Nullpointer suchen fuer Listenanzahl
        if ( *(uint32_t *)&m_fileAES.m_flash[pLB + i * sizeof( uint32_t )] == 0ul )
        {
            break;
        }
    }

    // Liste gefunden?
    if ( i > 0 )
    {
        // Pointer auf letzte definierte Liste
        pLB = *(uint32_t *)(&m_fileAES.m_flash[pLB + (i - 1) * sizeof( uint32_t )]);
        root().changeOrderInt32( (uint8_t *)&pLB );
        pLB -= FLASH_KONFIG_START;

        StructRecordDescription *pLBStruct = (StructRecordDescription *)(&m_fileAES.m_flash[pLB]);

        // Pointer auf Feldarray der Liste
        uint32_t pFeldArray = pLBStruct->ptr;
        root().changeOrderInt32( (uint8_t *)&pFeldArray );
        pFeldArray -= FLASH_KONFIG_START;

        // Felder vorhanden?
        if ( pLBStruct->fields )
        {
            // Pointer auf letztes Feld der Listenbeschreibung ermitteln.
            uint32_t pFeld = *(uint32_t *)&m_fileAES.m_flash[pFeldArray + (pLBStruct->fields - 1) * sizeof( uint32_t )];
            root().changeOrderInt32( (uint8_t *)&pFeld );
            pFeld -= FLASH_KONFIG_START;

            // Feldadresse + eigene Groesse bildet FlashSetupSize
            return pFeld + sizeof( StructRecordField );
        }


        // Keine Felder definiert!
        return pFeldArray;
    }

    // Keine Listenbeschreibung definiert!
    // Die Anzahl fuer Listenbeschreibungspointer muss mitberechnet werden,
    // um bei einer Uebertragung diese zu loeschen.
    return pLB + MAX_RECORD_DESCRIPTION * sizeof( uint32_t );
}

uint32_t DocumentStudio::GetFlashSetupSize4_2()
{
    return (EEP_SIZE_SU + sizeof( StructConfig ) + 0x1C + GetFlashSetupSize() + sizeof( StructFileStudioHeader ));
}

uint32_t DocumentStudio::GetMDEEKFlashSize()
{
    // Erweiterte Struktur vorhanden?
    if ( m_fileAES.m_extended_header.check != EXTENDED_CHECK )
    {
        return 0;
    }

    // Strukturen und Eingabeketten fuer MDE lesen.
    if ( m_fileAES.m_extended_header.offsetFreeScope == 1 )
    {
        // Kein Eintrag bei neuem Dokument. Erst nach writeDocument
        // 4Byte Anzahl Eingabeketten fuer MDE
        return 4;
    }
    else if ( m_fileAES.m_extended_header.offsetFreeScope != 0 )
    {
        // Nicht gerade mit DLL ausgelesen.
        // Groesse Eingabeketten fuer MDE
        return m_fileAES.m_extended_header.offsetAccessControlTablesV1 - m_fileAES.m_extended_header.offsetEventChains;
    }

    int32_t Unterpunkte;
    uint32_t pEK;

    Unterpunkte = m_fileAES.m_flashExtended[m_fileAES.m_extended_header.offsetEventChains];
    if ( Unterpunkte == 0 )
    {
        // 4Byte Anzahl Eingabeketten fuer MDE
        return 4;
    }

    // Pointer der letzten Eingabekette ermitteln
    pEK = m_fileAES.m_extended_header.offsetEventChains + sizeof( uint32_t ) + (Unterpunkte - 1) * sizeof( uint32_t );
    pEK = *(uint32_t *)&m_fileAES.m_flashExtended[pEK];
    root().changeOrderInt32( (uint8_t *)&pEK );
    // 16kByte fuer FlashOffset + Basisadresse fuer Konfigdaten subtrahieren
    pEK -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

    // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
    pEK += sizeof( StructOperating );
    StructReadRecord *pDE = (StructReadRecord *)&m_fileAES.m_flashExtended[m_fileAES.m_extended_header.offsetEventChains + pEK];

    // Pointer auf Feldarray der Eingabekette
    uint32_t pFeldArray = pDE->ptr;
    root().changeOrderInt32( (uint8_t *)&pFeldArray );
    pFeldArray -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

    // Felder vorhanden?
    if ( pDE->fields > 0 )
    {
        // Pointer auf letztes Feld der Eingabekette ermitteln.
        uint32_t pFeld = *(uint32_t *)&m_fileAES.m_flashExtended[m_fileAES.m_extended_header.offsetEventChains + pFeldArray + (pDE->fields - 1) * sizeof( uint32_t )];
        root().changeOrderInt32( (uint8_t *)&pFeld );
        pFeld -= (FLASH_SIZE_SU + FLASH_KONFIG_START);

        //StructReadField *pDF = (StructReadField*)&ExtendedMem[extBegin.offsetEventChains + pFeld];

        // Feldadresse + eigene Groesse bildet FlashSetupSize
        return pFeld + sizeof( StructReadField );
    }

    // Keine Eingabekettenfelder definiert!
    return pFeldArray;
}

uint32_t DocumentStudio::GetEntranceEKFlashSize()
{
    // Erweiterte Struktur vorhanden?
    if ( m_fileAES.m_extended_header.check != EXTENDED_CHECK )
    {
        return 0;
    }

    // Strukturen und Eingabeketten fuer Zutritt lesen.
    if ( m_fileAES.m_extended_header.offsetFreeScope == 1 )
    {
        // Kein Eintrag bei neuem Dokument. Erst nach writeDocument
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }
    else if ( m_fileAES.m_extended_header.offsetFreeScope != 0 )
    {
        // Nicht gerade mit DLL ausgelesen.
        // Groesse Eingabeketten fuer Zutritt
        return m_fileAES.m_extended_header.offsetFreeScope - m_fileAES.m_extended_header.offsetAccessControlEventChainsV1;
    }

    int32_t Unterpunkte;
    uint32_t pEK;

    Unterpunkte = m_fileAES.m_flashExtended[m_fileAES.m_extended_header.offsetAccessControlEventChainsV1];
    root().changeOrderInt32( (uint8_t *)&Unterpunkte );
    if ( Unterpunkte == 0 )
    {
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }

    // Pointer der letzten Eingabekette ermitteln
    pEK = m_fileAES.m_extended_header.offsetAccessControlEventChainsV1 + sizeof( uint32_t ) + (Unterpunkte - 1) * sizeof( uint32_t );
    pEK = *(uint32_t *)&m_fileAES.m_flashExtended[pEK];
    root().changeOrderInt32( (uint8_t *)&pEK );
    // Basisadresse fuer Konfigdaten subtrahieren
    pEK -= FLASH_KONFIG_START;

    // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
    pEK += sizeof( StructOperating );
    StructReadRecord *pDE = (StructReadRecord *)&m_fileAES.m_flashExtended[m_fileAES.m_extended_header.offsetAccessControlEventChainsV1 + pEK];

    // Pointer auf Feldarray der Eingabekette
    uint32_t pFeldArray = pDE->ptr;
    root().changeOrderInt32( (uint8_t *)&pFeldArray );
    pFeldArray -= FLASH_KONFIG_START;

    // Felder vorhanden?
    if ( pDE->fields > 0 )
    {
        // Pointer auf letztes Feld der Eingabekette ermitteln.
        uint32_t pFeld = *(uint32_t *)&m_fileAES.m_flashExtended[m_fileAES.m_extended_header.offsetAccessControlEventChainsV1 + pFeldArray + (pDE->fields - 1) * sizeof( uint32_t )];
        root().changeOrderInt32( (uint8_t *)&pFeld );
        pFeld -= FLASH_KONFIG_START;

        //StructReadField *pDF = (StructReadField*)&ExtendedMem[extBegin.offsetEventChains + pFeld];

        // Feldadresse + eigene Groesse bildet FlashSetupSize
        return pFeld + sizeof( StructReadField );
    }

    // Keine Eingabekettenfelder definiert!
    return pFeldArray;
}

uint32_t DocumentStudio::GetEntrance2EKFlashSize()
{
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash *)m_fileAES.m_flashExtended;

    // Erweiterte Struktur vorhanden?
    if ( extBegin.check != EXTENDED_CHECK )
    {
        return 0;
    }

    // Strukturen und Eingabeketten fuer Zutritt lesen.
    if ( extBegin.offsetFreeScope == 1 )
    {
        // Kein Eintrag bei neuem Dokument. Erst nach writeDocument
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }
    /*else if (extBegin.offsetFreeScope != 0) {

    // Nicht gerade mit DLL ausgelesen.
    // Groesse Eingabeketten fuer Zutritt
    return extBegin.offsetFreeScope - extBegin.offsetAccessControlEventChainsV2;
    }*/

    int32_t Unterpunkte = 0;
    uint32_t pEK;

    if ( extBegin.offsetAccessControlEventChainsV2 != 0 )
    {
        Unterpunkte = *(uint32_t *)&m_fileAES.m_flashExtended[extBegin.offsetAccessControlEventChainsV2];
    }

    if ( Unterpunkte == 0 )
    {
        // 4Byte Anzahl Eingabeketten fuer Zutritt
        return 4;
    }

    // Pointer der letzten Eingabekette ermitteln
    pEK = extBegin.offsetAccessControlEventChainsV2 + sizeof( uint32_t ) + (Unterpunkte - 1) * sizeof( uint32_t );
    pEK = *(uint32_t *)&m_fileAES.m_flashExtended[pEK];
    //  root().ChangeOrderLong((uint8_t*)&pEK);
    // Basisadresse fuer Konfigdaten subtrahieren
    pEK -= FLASH_KONFIG_START;

    // vor jeder Eingabekettenstruktur liegt eine Bedien - Struktur
    pEK += sizeof( StructOperating );
    StructReadRecord *pDE = (StructReadRecord *)&m_fileAES.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pEK];

    // Pointer auf Feldarray der Eingabekette
    uint32_t pFeldArray = pDE->ptr;
    root().changeOrderInt32( (uint8_t *)&pFeldArray );
    pFeldArray -= FLASH_KONFIG_START;

    // Felder vorhanden?
    if ( pDE->fields > 0 )
    {
        // Pointer auf letztes Feld der Eingabekette ermitteln.
        uint32_t pFeld = *(uint32_t *)&m_fileAES.m_flashExtended[extBegin.offsetAccessControlEventChainsV2 + pFeldArray + (pDE->fields - 1) * sizeof( uint32_t )];
        root().changeOrderInt32( (uint8_t *)&pFeld );
        pFeld -= FLASH_KONFIG_START;

        //StructReadField *pDF = (StructReadField*)&ExtendedMem[extBegin.offsetEventChains + pFeld];

        // Feldadresse + eigene Groesse bildet FlashSetupSize
        return pFeld + sizeof( StructReadField );
    }

    // Keine Eingabekettenfelder definiert!
    return pFeldArray;
}

uint32_t DocumentStudio::GetExtendedFlashSize()
{
    // Erweiterte Struktur vorhanden?
    if ( m_fileAES.m_extended_header.check != EXTENDED_CHECK )
    {
        return 0;
    }

    return m_fileAES.m_extended_header.offsetFreeScope;
}

bool DocumentStudio::IsEmpty()
{
    // Das Dokument gilt als Leer, wenn kein Menueeintrag vorhanden ist.
    EntryDevice &device = root();
    EntryHandling &handling = rootHandling();
    EntryMenu *pMenu;

    // Kein Hauptmenue, dann ist Dokument leer
    if ( handling.size() == 0 )
    {
        // Bei folgenden Setuptypen ist kein Hauptmenü nötig!
        if ( device.hasSetupType( {
                SetupType::ZK_MasterIV,
                SetupType::Mobil_MasterIV,
                SetupType::MDE_BoxIV,
                SetupType::IPC_Extension,
                SetupType::IO_Box,
                SetupType::KYO_Inloc,
                SetupType::EVO_3_5_Universal,
                SetupType::Mobil_Box,
                SetupType::Timeboy_TBDock,
                SetupType::Mobil_TBDock,
                SetupType::KYO_Oneloc,
                SetupType::KYO_Cenloc } ) )
        {
            return false;
        }

        return true;
    }

    pMenu = (EntryMenu *)handling.first();

    // Keine Eintrage, dann ist Dokument leer
    if ( pMenu->size() == 0 )
    {
        return true;
    }

    return false;
}

EntryDevice &DocumentStudio::root()
{
    // Pointer auf Dokumentwurzel holen
    return *m_root;
}

EntryTableRecord &DocumentStudio::rootTableRecord()
{
    return root().rootTableRecord();
}

EntryTableList &DocumentStudio::rootTableList()
{
    return root().rootTableList();
}

EntryHandling &DocumentStudio::rootHandling()
{
    return root().rootHandling();
}

EntrySignalProcessing &DocumentStudio::rootSignalProcessing()
{
    return root().rootSignalProcessing();
}

EntryEventChains &DocumentStudio::rootEventChains()
{
    return root().rootEventChains();
}

EntryIOModule &DocumentStudio::rootIOModule()
{
    return root().rootIOModule();
}

EntryAccessControl2 &DocumentStudio::rootAccessControl()
{
    return root().rootAccessControl();
}

EntryEnergieSchemes &DocumentStudio::rootEnergieSchemes()
{
    return root().rootEnergieSchemes();
}

const char *DocumentStudio::setupPassword()
{
    memset( m_setupPassword, 0, sizeof( m_setupPassword ) );
    if ( strlen( (char *)root().m_extended.passwordSetup ) )
    {
        // Das Passwort des Eintrags liegt verschlüsselt vor und muss deswegen
        // über einen localen puffer entschlüsselt und bereitgestellt werden.
        memcpy( m_setupPassword, (char *)root().m_extended.passwordSetup, sizeof( m_setupPassword ) );
#ifdef DATAFOX_STUDIO_IV
        df4_crypt_internal df4;
        df4.crypt( (uint8_t *)m_setupPassword, sizeof( m_setupPassword ) );
#endif
        return (const char *)m_setupPassword;
    }

    return nullptr;
}

std::string DocumentStudio::fileNameList( int32_t num )
{
    assert( num < MAX_RECORD_DESCRIPTION );
    return Utils::File::replaceDirSeparator( m_fileAES.m_listFileName[num] );
}

void DocumentStudio::setFileNameList( int32_t num, const std::string &fileName )
{
    assert( num < MAX_RECORD_DESCRIPTION );
    m_fileAES.m_listFileName[num] = fileName;
    Utils::File::replaceDirSeparator( m_fileAES.m_listFileName[num] );
}

std::string DocumentStudio::fileNameListAccessControl( int32_t num )
{
    assert( num < MAX_RECORD_DESCRIPTION );
    return Utils::File::replaceDirSeparator( m_fileAES.m_listEntrance2FileName[num] );
}

void DocumentStudio::setFileNameListAccessControl( int32_t num, const std::string &fileName )
{
    assert( num < MAX_RECORD_DESCRIPTION );
    m_fileAES.m_listEntrance2FileName[num] = fileName;
    Utils::File::replaceDirSeparator( m_fileAES.m_listEntrance2FileName[num] );
}

std::string DocumentStudio::fileNameExtended( int32_t num )
{
    assert( num < MAX_RECORD_DESCRIPTION );
    return Utils::File::replaceDirSeparator( m_fileAES.m_strExtendedFilePath[num] );
}

void DocumentStudio::setFileNameExtended( int32_t num, const std::string &fileName )
{
    assert( num < MAX_RECORD_DESCRIPTION );
    m_fileAES.m_strExtendedFilePath[num] = fileName;
    Utils::File::replaceDirSeparator( m_fileAES.m_strExtendedFilePath[num] );
}

bool DocumentStudio::isAccessControl()
{
    return (root().m_config.enableEntrance2);
}

bool DocumentStudio::isAccessControlExtension()
{
    return (root().m_extended.acExtension);
}

void DocumentStudio::setVersions( const std::string &studioVersion, const std::string &libraryVersion )
{
    // Aktuelle Versionen in Dateiheader eintragen
    m_fileAES.SetStudioVersion( studioVersion.c_str() );
    m_fileAES.SetLibraryVersion( libraryVersion.c_str() );

    m_fileXML.setVersions( studioVersion, libraryVersion );
}

void DocumentStudio::setVersions( uint32_t studioVersion, uint32_t libraryVersion )
{
    setVersions( versionFormat( studioVersion ), versionFormat( libraryVersion ) );
}

std::string DocumentStudio::studioVersion()
{
    return versionFormat( m_fileAES.m_file_header.studioVersion );
}

std::string DocumentStudio::minStudioVersion()
{
    return versionFormat( m_fileAES.m_file_header.minStudioVersion );
}

std::string DocumentStudio::shouldStudioVersion()
{
    return versionFormat( m_fileAES.m_file_header.shouldStudioVersion );
}

std::string DocumentStudio::minLibraryVersion()
{
    return versionFormat( m_fileAES.m_file_header.minLibraryVersion );
}

std::string DocumentStudio::shouldLibraryVersion()
{
    return versionFormat( m_fileAES.m_file_header.shouldLibraryVersion );
}

std::string DocumentStudio::minFirmware1Version()
{
    return versionFormat( m_fileAES.m_file_header.minFirmware1Version );
}

std::string DocumentStudio::shouldFirmware1Version()
{
    return versionFormat( m_fileAES.m_file_header.shouldFirmware1Version );
}

bool DocumentStudio::supportFirmware1Version()
{
    return m_fileAES.supportFirmware1Version();
}

std::string DocumentStudio::minFirmware2Version()
{
    return versionFormat( m_fileAES.m_file_header.minFirmware2Version );
}

std::string DocumentStudio::shouldFirmware2Version()
{
    return versionFormat( m_fileAES.m_file_header.shouldFirmware2Version );
}

bool DocumentStudio::supportFirmware2Version()
{
    return m_fileAES.supportFirmware2Version();
}

std::string DocumentStudio::minFirmware3Version()
{
    return versionFormat( m_fileAES.m_file_header.minFirmware3Version );
}

std::string DocumentStudio::shouldFirmware3Version()
{
    return versionFormat( m_fileAES.m_file_header.shouldFirmware3Version );
}

bool DocumentStudio::supportFirmware3Version()
{
    return m_fileAES.supportFirmware3Version();
}

std::string DocumentStudio::versionFormat( uint32_t version )
{
    std::string res;

    if ( version )
    {
        if ( !(FileAES::is_invalid_header_version( version )) )
        {
            res = boost::str( boost::format( "%1$02d.%2$02d.%3$02d.%4$02d" ) % ((version >> 24) & 0xFF) % ((version >> 16) & 0xFF) % ((version >> 8) & 0xFF) % (version & 0xFF) );
        }
        else
        {
            res = "UNSUPPORTED";
        }
    }

    return res;
}

std::list<std::string> DocumentStudio::globalVariablesName()
{
    std::list<std::string> res;
    int32_t i;

    for ( i = 0; i < 8; i++ )
    {
        if ( ((StructGlobalVar *)m_fileAES.m_flash)->names[i][0] )
        {
            res.push_back( ((StructGlobalVar *)m_fileAES.m_flash)->names[i] );
        }
        else
        {
            res.push_back( "" );
        }
    }

    return res;
}

std::unordered_map<std::string, std::string> DocumentStudio::globalVariablesValueDefault()
{
    int32_t i;
    std::unordered_map<std::string, std::string> res;
    StructExtendedFlash extBegin = *(StructExtendedFlash *)m_fileAES.m_flashExtended;

    // GV setzen?
    if ( extBegin.offsetGlobalVariableValues )
    {
        // Struktur lesen
        StructValueGlobalVar ValueGVx = *(StructValueGlobalVar *)&m_fileAES.m_flashExtended[extBegin.offsetGlobalVariableValues];

        for ( i = 0; i < 8; i++ )
        {
            if ( ((StructGlobalVar *)m_fileAES.m_flash)->names[i][0] && (ValueGVx.flags & (1 << i)) == (1 << i) )
            {
                res[((StructGlobalVar *)m_fileAES.m_flash)->names[i]] = ValueGVx.value[i];
            }
        }
    }

    return res;
}
