// file_studio_setup_xml.cpp: Implementierung der Klasse CDFFileStudioXML.
//
//////////////////////////////////////////////////////////////////////

#ifndef TIXML_USE_STL
#define TIXML_USE_STL
#endif
#include "file/file_studio_setup_xml.h"
#include <map>

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

CDFFileStudioXML::CDFFileStudioXML()
{
    // Member initialisieren
    Reset();

    // Mapping von Gerätetyp auf Setuptyp
    m_setupType[0]  = SetupType::PZE_MasterIV; // IV_VALUE_TYPE_DEVICE_PZE
    m_setupType[1]  = SetupType::ZK_MasterIV; // IV_VALUE_TYPE_DEVICE_ZK
    m_setupType[2]  = SetupType::BDE_MasterIV; // IV_VALUE_TYPE_DEVICE_BDE
    m_setupType[3]  = SetupType::Mobil_MasterIV; // IV_VALUE_TYPE_DEVICE_MOBIL
    m_setupType[4]  = SetupType::MDE_BoxIV; // IV_VALUE_TYPE_DEVICE_MDE
    m_setupType[5]  = SetupType::AE_MasterIV ; // IV_VALUE_TYPE_DEVICE_AE
    m_setupType[6]  = SetupType::Timeboy_MasterIV; // IV_VALUE_TYPE_DEVICE_TIMEBOY
    m_setupType[7]  = SetupType::Exklusiv_LineIV; // IV_VALUE_TYPE_DEVICE_EXKLUSIV
    m_setupType[8]  = SetupType::Flex_MasterIV; // IV_VALUE_TYPE_DEVICE_FLEX
    m_setupType[9]  = SetupType::Timeboy_Mobil_PZE; // IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE
    m_setupType[10] = SetupType::EVO_2_8_Pure; // IV_VALUE_TYPE_DEVICE_EVO_2_8_PURE
    m_setupType[11] = SetupType::EVO_4_3; // IV_VALUE_TYPE_DEVICE_EVO_4_3
    m_setupType[12] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7
    m_setupType[13] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10
    m_setupType[14] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7
    m_setupType[15] = SetupType::IO_Box; // IV_VALUE_TYPE_DEVICE_IO_BOX
    m_setupType[16] = SetupType::KYO_Inloc; // IV_VALUE_TYPE_DEVICE_KYO_INLOC
    m_setupType[17] = SetupType::Mobil_Box; // IV_VALUE_TYPE_DEVICE_MOBIL_BOX
    m_setupType[18] = SetupType::Timeboy_TBDock; // IV_VALUE_TYPE_DEVICE_TIMEBOY_TBDOCK
    m_setupType[19] = SetupType::Mobil_TBDock; // IV_VALUE_TYPE_DEVICE_MOBIL_TBDOCK
    m_setupType[20] = SetupType::KYO_Oneloc; // IV_VALUE_TYPE_DEVICE_KYO_ONELOC
    m_setupType[21] = SetupType::EVO_3_5_Universal; // IV_VALUE_TYPE_DEVICE_EVO_3_5_UNIVERSAL
    //m_setupType[22] = ; // IV_VALUE_TYPE_DEVICE_SAMS_OCL_READER
    m_setupType[23] = SetupType::EVO_2_8_Pure; // IV_VALUE_TYPE_DEVICE_EVO_3_5_PURE
    //m_setupType[24] = ; // IV_VALUE_TYPE_DEVICE_EVO_AGERA
    m_setupType[25] = SetupType::KYO_Cenloc; // IV_VALUE_TYPE_DEVICE_KYO_CENLOC
    m_setupType[26] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_VARIO_10
    m_setupType[27] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_VARIO_12
    m_setupType[28] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_VARIO_15
    m_setupType[29] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_VARIO_17
    m_setupType[30] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_IPC_EVO_12_1
    m_setupType[31] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_IPC_EVO_15
    m_setupType[32] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_IPC_EVO_18_5
    m_setupType[33] = SetupType::IPC_Extension; // IV_VALUE_TYPE_DEVICE_IPC_EVO_24
    //m_setupType[34] = ; // IV_VALUE_TYPE_DEVICE_EVO_INTERA_2
    m_setupType[35] = SetupType::EVO_4_3; // IV_VALUE_TYPE_DEVICE_EVO_4_6_FLEX_KEY
    m_setupType[36] = SetupType::EVO_2_8_Pure; // IV_VALUE_TYPE_DEVICE_EVO_5_0_PURE
}

CDFFileStudioXML::~CDFFileStudioXML()
{

}

void CDFFileStudioXML::Reset()
{
    // Member initialisieren
    m_errorText.clear();

    // Dokument leeren
    m_doc.Clear();
}

int CDFFileStudioXML::Modify( const char *data, int *piError )
{
    tinyxml2::XMLDocument doc;

    // Handelt es sich um Daten oder eine Dateiangabe?
    if ( strstr( data, "<?xml" ) == data )
    {
        // Versuchtes Parsen der Daten.
        if ( doc.Parse( data ) != tinyxml2::XML_SUCCESS )
        {
            *piError = IV_ERROR_FILE_LOAD_XML;
            return 0;
        }
    }
    else
    {
        // Versuchtes Laden einer Datei.
        if ( doc.LoadFile( data ) != tinyxml2::XML_SUCCESS )
        {
            *piError = IV_ERROR_FILE_LOAD_XML;
            return 0;
        }
    }

    // Die gewuenschten Aenderungen auswerten.
    tinyxml2::XMLElement *element = doc.RootElement();
    if ( strcmp( element->Value(), "modify" ) != 0 )
    {  
        *piError = IV_ERROR_FILE_CONTENT_XML;
        return 0;
    }


    int res;
    CModifyData modifyData;
    modifyData.Reset();

    // Grunddaten zuweisen
    modifyData.file = this;

    if ( (res = Extract( (char*)"", element, modifyData )) != IV_ERROR_SUCCESS )
    {
        // Fehlercode zurueckgeben
        *piError = res;
        return 0;
    }

    // Dokument initialisieren
    Reset();

    // Rooteintrag erzeugen
    CDFItemDevice *item = new CDFItemDevice( &m_doc );
    m_doc.InsertEndChild( item->xmlElement() );

    // Die gewuenschten Aenderungen vornehmen.
    if ( !item->Modify( modifyData ) )
    {
        // Fehler bei Uebernahme der Modifizierungsdaten
        *piError = IV_ERROR_FILE_MODIFY;
        return 0;
    }

    return 1;
}

int CDFFileStudioXML::Extract( const char *elementPathParam, tinyxml2::XMLElement *element, CModifyData &modifyData )
{
    static int level = 0;
    static StructModifyDataPassword password;
    static StructModifyDataReplaceString replaceString;

    // Initialisieren der Strukturen
    password.password = -1;

    // Pfadangabe zu Element
    char elementPath[128];
    memset( elementPath, 0, sizeof(elementPath) );
    sprintf( elementPath, "%s/%s", elementPathParam, element->Value() );

    level++; // Baumtiefe halten

    if ( level > 4 )
    {
        level--;
        return IV_ERROR_FILE_CONTENT_XML;
    }

    // Wert ermitteln
    if ( element->NoChildren() )
    {
        level--;
        return IV_ERROR_FILE_CONTENT_XML;
    }

    //###########################################
    // Beginn: Extrahieren der XML-Daten
    if ( strcmp( elementPath, "/modify" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/device" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/device/type" ) == 0 )
    {
        std::string deviceTypeStr = element->FirstChild()->Value();
        int deviceType = -1;

        if ( deviceTypeStr == "PZE-MasterIV" )
            deviceType = 0; // IV_VALUE_TYPE_DEVICE_PZE;
        else if (deviceTypeStr == "ZK-MasterIV")
            deviceType = 1; // IV_VALUE_TYPE_DEVICE_ZK;
        else if (deviceTypeStr == "BDE-MasterIV")
            deviceType = 2; // IV_VALUE_TYPE_DEVICE_BDE;
        else if (deviceTypeStr == "Mobil-MasterIV" || deviceTypeStr == "MOBIL-MasterIV")
            deviceType = 3; // IV_VALUE_TYPE_DEVICE_MOBIL;
        else if (deviceTypeStr == "MDE-BoxIV" || deviceTypeStr == "MDE-MasterIV")
            deviceType = 4; // IV_VALUE_TYPE_DEVICE_MDE;
        else if (deviceTypeStr == "AE-MasterIV")
            deviceType = 5; // IV_VALUE_TYPE_DEVICE_AE;
        else if (deviceTypeStr == "TimeboyIV")
            deviceType = 6; // IV_VALUE_TYPE_DEVICE_TIMEBOY
        else if (deviceTypeStr == "Exklusiv-Line IV")
            deviceType = 7; // IV_VALUE_TYPE_DEVICE_EXKLUSIV
        else if (deviceTypeStr == "Flex-MasterIV")
            deviceType = 8; // IV_VALUE_TYPE_DEVICE_FLEX
        else if (deviceTypeStr == "TimeboyIV Mobil-PZE" || deviceTypeStr == "Timeboy-Mobil-PZE")
            deviceType = 9; // IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE
        else if (deviceTypeStr == "EVO 2.8 Pure" || deviceTypeStr == "EVO-Line 2.8")
            deviceType = 10; // IV_VALUE_TYPE_DEVICE_TERMINAL_2_8
        else if (deviceTypeStr == "EVO 4.3" || deviceTypeStr == "EVO-Line 4.3")
            deviceType = 11; // IV_VALUE_TYPE_DEVICE_TERMINAL_4_3
        else if (deviceTypeStr == "IPC-Extension A")
            deviceType = 12; // IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_5_7
        else if (deviceTypeStr == "IPC-Extension B")
            deviceType = 13; // IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_VARIO_10
        else if (deviceTypeStr == "IPC-Extension C")
            deviceType = 14; // IV_VALUE_TYPE_DEVICE_IPC_EXTENSION_Q7
        else if (deviceTypeStr == "IO-Box V4")
            deviceType = 15; // IV_VALUE_TYPE_DEVICE_IO_BOX
        else if (deviceTypeStr == "KYO Inloc" || deviceTypeStr == "ZK-Box V4")
            deviceType = 16; // IV_VALUE_TYPE_DEVICE_KYO_INLOC
        else if (deviceTypeStr == "Mobil-Box V4")
            deviceType = 17; // IV_VALUE_TYPE_DEVICE_MOBIL_BOX
        else if (deviceTypeStr == "Dockingstation V2")
            deviceType = 18; // IV_VALUE_TYPE_DEVICE_TIMEBOY_TBDOCK
        else if (deviceTypeStr == "Fahrzeugdatenlogger V2")
            deviceType = 19; // IV_VALUE_TYPE_DEVICE_MOBIL_TBDOCK
        else if (deviceTypeStr == "KYO Oneloc" || deviceTypeStr == "ZK-Knoten")
            deviceType = 20; // IV_VALUE_TYPE_DEVICE_KYO_ONELOC
        else if (deviceTypeStr == "EVO 3.5 Universal")
            deviceType = 21; // IV_VALUE_TYPE_DEVICE_EVO_3_5 UNIVERSAL
        else if (deviceTypeStr == "SAMS-OCL-Reader")
            deviceType = 22; // IV_VALUE_TYPE_DEVICE_SAMS_OCL_READER
        else if (deviceTypeStr == "EVO 3.5 Pure" || deviceTypeStr == "EVO 3.5")
            deviceType = 23; // IV_VALUE_TYPE_DEVICE_TERMINAL_3_5
        else if (deviceTypeStr == "EVO Agera" || deviceTypeStr == "ZK-Leser Plus")
            deviceType = 24; // IV_VALUE_TYPE_DEVICE_EVO_AGERA
        else if (deviceTypeStr == "KYO Cenloc" || deviceTypeStr == "ZK-Zentrale")
            deviceType = 25; // IV_VALUE_TYPE_DEVICE_KYO_CENLOC
        else if ( deviceTypeStr == "Vario 10" )
            deviceType = 26; // IV_VALUE_TYPE_DEVICE_VARIO_10
        else if ( deviceTypeStr == "Vario 12" )
            deviceType = 27; // IV_VALUE_TYPE_DEVICE_VARIO_12
        else if ( deviceTypeStr == "Vario 15" )
            deviceType = 28; // IV_VALUE_TYPE_DEVICE_VARIO_15
        else if ( deviceTypeStr == "Vario 17" )
            deviceType = 29; // IV_VALUE_TYPE_DEVICE_VARIO_17
        else if ( deviceTypeStr == "IPC EVO 12.1" )
            deviceType = 30; // IV_VALUE_TYPE_DEVICE_IPC_EVO_12_1
        else if ( deviceTypeStr == "IPC EVO 15" )
            deviceType = 31; // IV_VALUE_TYPE_DEVICE_IPC_EVO_15
        else if ( deviceTypeStr == "IPC EVO 18.5" )
            deviceType = 32; // IV_VALUE_TYPE_DEVICE_IPC_EVO_18_5
        else if ( deviceTypeStr == "IPC EVO 24" )
            deviceType = 33; // IV_VALUE_TYPE_DEVICE_IPC_EVO_24
        else if ( deviceTypeStr == "EVO Intera 2" )
            deviceType = 34; // IV_VALUE_TYPE_DEVICE_EVO_INTERA_2
        else if ( deviceTypeStr == "EVO 4.6 FlexKey" )
            deviceType = 35; // IV_VALUE_TYPE_DEVICE_EVO_4_6_FLEX_KEY
        else if ( deviceTypeStr == "EVO 5.0 Pure" )
            deviceType = 36; // IV_VALUE_TYPE_DEVICE_EVO_5_0_PURE

        // Geraetetyp unbekannt? Oder hat keinen zugewiesenen Setuptyp?
        if ( deviceType == -1 || !m_setupType.count( deviceType ) )
        {
            return IV_ERROR_FILE_CONTENT_XML;
        }

        // Wert uebernehmen
        modifyData.deviceType = deviceType;
        modifyData.setupType = m_setupType[deviceType];
        modifyData.flags |= SMDF_DEVICE_TYPE;

    } else if (strcmp(elementPath, "/modify/transponder") == 0) {

    } else if (strcmp(elementPath, "/modify/transponder/technic") == 0) {

        std::string transponderTypeStr = element->FirstChild()->Value();
        int transponderType = -1;

        if ( transponderTypeStr == "EM4102" )
            transponderType = TRANSPONDER_UNIQUE; // Unique
        else if ( transponderTypeStr == "HITAG-2" )
            transponderType = TRANSPONDER_HITAG2; // Hitag2
        else if ( transponderTypeStr == "Legic-Prime" )
            transponderType = TRANSPONDER_LEGIC; // Legic-Prime
        else if ( transponderTypeStr == "MIFARE" )
            transponderType = TRANSPONDER_MIFARE; // Mifare Classic
        else if ( transponderTypeStr == "iButton" )
            transponderType = TRANSPONDER_IBUTTON; // iButton
        else if ( transponderTypeStr == "SimonsVoss" )
            transponderType = TRANSPONDER_SIMONSVOSS; // SimonsVoss G1 / G2
        else if ( transponderTypeStr == "Nedap" )
            transponderType = TRANSPONDER_NEDAP; // Nedap
        else if ( transponderTypeStr == "HID-PROXKEY" )
            transponderType = TRANSPONDER_HID_PROXKEY; // HID-ProxKey
        else if ( transponderTypeStr == "HID-ICLASS" )
            transponderType = TRANSPONDER_HID_ICLASS; // HID-iClass
        else if ( transponderTypeStr == "WIEGAND" )
            transponderType = TRANSPONDER_WIEGAND; // Wiegand
        else if ( transponderTypeStr == "CLOCK_DATA" )
            transponderType = TRANSPONDER_CLOCK_DATA; // Clock Data
        else if ( transponderTypeStr == "ETSI" )
            transponderType = TRANSPONDER_BLUEBOX_ID_TRONIC; // ETSI
        else if ( transponderTypeStr == "NONE" )
            transponderType = TRANSPONDER_NONE; // Keinen

        // Transpondertyp unbekannt?
        if ( transponderType == -1 )
        {
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }

        // Wert uebernehmen
        modifyData.technic = transponderType;
        modifyData.flags |= SMDF_TRANSPONDER_TECHNIC;
    }
    else if ( strcmp( elementPath, "/modify/transponder/password" ) == 0 )
    {  
        // Technic muss bereits bekannt sein und Hitag2!
        if ( !(modifyData.flags & SMDF_TRANSPONDER_TECHNIC) || modifyData.technic != TRANSPONDER_HITAG2 )
        {
            // Undefinierter Transponder
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }

        size_t i;
        std::string pass = element->FirstChild()->Value();
        for ( i = 0; i < strlen( pass.c_str() ); ++i )
        {
            if ( !isxdigit( pass[i] ) )
                break;
        }

        if ( strlen( pass.c_str() ) != (sizeof(modifyData.hitagPassword) - 1) || (i < strlen( pass.c_str() )) )
        {
            // Ueberlanger Wert
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }

        // Passwort uebernehmen
        strcpy( modifyData.hitagPassword, pass.c_str() );
        modifyData.flagsTranponder |= TMDF_PASSWORD;
    }
    else if ( strcmp( elementPath, "/modify/transponder/inc_value" ) == 0 )
    {
        // Technic muss bereits bekannt sein!
        if ( !(modifyData.flags & SMDF_TRANSPONDER_TECHNIC) )
        {
            // Undefinierter Transponder
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }

        switch ( modifyData.technic )
        {
            case TRANSPONDER_HITAG2:  // Hitag2
            case TRANSPONDER_MIFARE:  // Mifare
            {
                modifyData.hitagAutoincValue = atoi( element->FirstChild()->Value() );
                if ( modifyData.hitagAutoincValue == 0 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Segment angegeben
                modifyData.flagsTranponder |= TMDF_AUTOINC_VALUE;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_2" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_3" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_1" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_2" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_3" ) == 0 )
    {

        // Technic muss bereits bekannt sein!
        if ( !(modifyData.flags & SMDF_TRANSPONDER_TECHNIC) )
        {
            // Undefinierter Transponder
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/serial_number" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_MIFARE: // Mifare
            {
                // Nur bei technic Mifare
                modifyData.serial_number = atoi( element->FirstChild()->Value() );
                if ( modifyData.serial_number != 0 && modifyData.serial_number != 1 )
                {
                    // Undefinierter Wert
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Seriennummer von Ausweis lesen
                modifyData.serial_number = (modifyData.serial_number == 1);
                modifyData.flagsTranponder |= TMDF_SERIAL_NUMBER;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/bit" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_UNIQUE: // Unique
            {
                // Nur bei technic Unique
                modifyData.unique = atoi( element->FirstChild()->Value() );
                if ( modifyData.unique != 32 && modifyData.unique != 40 )
                {
                    // Undefinierter Bitanzahl
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Bitanzahl angegeben
                modifyData.unique = (modifyData.unique == 40);
                modifyData.flagsTranponder |= TMDF_UNIQUE;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/type" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_HID_PROXKEY:
            case TRANSPONDER_HID_ICLASS: 
            {
                // Nur bei technic Unique
                unsigned char val = atoi(element->FirstChild()->Value());
                if ( val > 3 )
                {
                    // Undefinierter Bitanzahl
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Wert uebernehmen
                modifyData.transponderHID.type = val;
                // Typen angeben
                modifyData.flagsTranponder |= TMDF_HID_TYPE;
                break;
            }
          default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/segment" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_2/segment" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_3/segment" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_HITAG2: // Hitag2
            {
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( val != 0 && !(val >= 4 && val <= 7) )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                if ( strstr( elementPath, "valueblock_1" ) )
                {
                    // Segment angegeben
                    modifyData.transponder.tpSegment1 = val;
                    modifyData.flagsTranponder |= TMDF_SEGMENT1;
                }
                else if ( strstr( elementPath, "valueblock_2" ) )
                {
                    // Segment angegeben
                    modifyData.transponder.tpSegment2 = val;
                    modifyData.flagsTranponder |= TMDF_SEGMENT2;
                }
                else
                {
                    // Segment angegeben
                    modifyData.transponder.tpSegment3 = val;
                    modifyData.flagsTranponder |= TMDF_SEGMENT3;
                }
                break;
            }
            case TRANSPONDER_MIFARE: // Mifare
            {          
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( strstr( elementPath, "valueblock_1" ) == NULL || val > 15 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Segment angegeben
                modifyData.transponderMifare.sectorBlock |= (val & 0x0F) << 4;
                modifyData.flagsTranponder |= TMDF_SECTOR;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/block" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_MIFARE: // Mifare
            {
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( val > 2 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Block angegeben
                modifyData.transponderMifare.sectorBlock |= (val & 0x0F);
                modifyData.flagsTranponder |= TMDF_BLOCK;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/format" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_2/format" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_3/format" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_UNIQUE: // Unique
            {
                // Nur bei technic Unique
                modifyData.uniqueMode = atoi( element->FirstChild()->Value() );
                if ( modifyData.uniqueMode != 0 && modifyData.uniqueMode != 1 && modifyData.uniqueMode != 3 )
                {
                    // Undefiniertes Format
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Wert liegt an 0x30 im Byte
                modifyData.uniqueMode <<= 4;

                // Format angegeben
                modifyData.flagsTranponder |= TMDF_UNIQUE_MODE;
                break;
            }
            case TRANSPONDER_HITAG2: // Hitag2
            {
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( val > 3 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                if ( strstr( elementPath, "valueblock_1" ) )
                {
                    // Modusvariable ggf. initialisieren
                    if ( !(modifyData.flagsTranponder & TMDF_MODE1_LENGTH) )
                    {
                        modifyData.transponder.tpMode1 = 0x80;
                    }
                    // Format in Modusvariable uebernehmen
                    modifyData.transponder.tpMode1 |= val << 4;
                    // Format angegeben
                    modifyData.flagsTranponder |= TMDF_MODE1_FORMAT;
                }
                else if ( strstr( elementPath, "valueblock_2" ) )
                {
                    // Format in Modusvariable uebernehmen
                    modifyData.transponder.tpMode2 |= val << 4;
                    // Format angegeben
                    modifyData.flagsTranponder |= TMDF_MODE2_FORMAT;
                }
                else
                {
                    // Format in Modusvariable uebernehmen
                    modifyData.transponder.tpMode3 |= val << 4;
                    // Format angegeben
                    modifyData.flagsTranponder |= TMDF_MODE3_FORMAT;
                }

                break;
            }
            case TRANSPONDER_MIFARE: // Mifare
            {
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( strstr( elementPath, "valueblock_1" ) == NULL || val > 3 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Modusvariable ggf. initialisieren
                if ( !(modifyData.flagsTranponder & TMDF_MODE_LENGTH) )
                {
                    modifyData.transponderMifare.mode = BDTP_ACTIVE;
                }

                // Format in Modusvariable uebernehmen
                modifyData.transponderMifare.mode |= (val & 0x03) << 4;
                // Format angegeben
                modifyData.flagsTranponder |= TMDF_MODE_FORMAT;
                break;
            }
            case TRANSPONDER_HID_PROXKEY: // HID-Proxkey
            case TRANSPONDER_HID_ICLASS: // HID-iClass
            {
                // Nur bei technic HID
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( strstr( elementPath, "valueblock_1" ) != NULL && val != 0 && val != 1 && val != 3 )
                {
                    // Undefiniertes Format
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Wert uebernehmen
                modifyData.transponderHID.format = val;
                // Format angegeben
                modifyData.flagsTranponder |= TMDF_HID_FORMAT;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/position" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_MIFARE: // Mifare
            {
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( strstr( elementPath, "valueblock_1") == NULL || val > 14 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Startposition uebernehmen
                modifyData.transponderMifare.cut |= (val & 0x0F) << 4;
                modifyData.flagsTranponder |= TMDF_POSITION;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    } 
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/length" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_2/length" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_3/length" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_HITAG2: // Hitag2 
            {

                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( val > 10 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                if ( strstr( elementPath, "valueblock_1" ) )
                {
                    // Modusvariable ggf. initialisieren
                    if ( !(modifyData.flagsTranponder & TMDF_MODE1_FORMAT) )
                    {
                        modifyData.transponder.tpMode1 = 0x80;
                    }
                    // Laenge in Modusvariable uebernehmen
                    modifyData.transponder.tpMode1 |= val;
                    // Laenge angegeben
                    modifyData.flagsTranponder |= TMDF_MODE1_LENGTH;
                }
                else if ( strstr( elementPath, "valueblock_2" ) )
                {
                    // Laenge in Modusvariable uebernehmen
                    modifyData.transponder.tpMode2 |= val;
                    // Laenge angegeben
                    modifyData.flagsTranponder |= TMDF_MODE2_LENGTH;
                }
                else
                {
                    // Laenge in Modusvariable uebernehmen
                    modifyData.transponder.tpMode3 |= val;
                    // Laenge angegeben
                    modifyData.flagsTranponder |= TMDF_MODE3_LENGTH;
                }
                break;
            }
            case TRANSPONDER_MIFARE: // Mifare 
            {
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( strstr( elementPath, "valueblock_1" ) == NULL || val > 15 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Modusvariable ggf. initialisieren
                if ( !(modifyData.flagsTranponder & TMDF_MODE_FORMAT) )
                {
                    modifyData.transponderMifare.mode = BDTP_ACTIVE;
                }
                // Laenge in Modusvariable uebernehmen
                modifyData.transponderMifare.mode |= (val & 0x0F);
                // Laenge angegeben
                modifyData.flagsTranponder |= TMDF_MODE_LENGTH;
                break;
            }
            case TRANSPONDER_HID_PROXKEY: // HID-Proxkey
            case TRANSPONDER_HID_ICLASS: // HID-iClass
            {
                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( strstr( elementPath, "valueblock_1" ) == NULL || val > 12 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Wert uebernehmen.
                modifyData.transponderHID.length = val;
                // Laenge angegeben
                modifyData.flagsTranponder |= TMDF_HID_LENGTH;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/init_value" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_2/init_value" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_3/init_value" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_HITAG2: // Hitag2 
            {

                unsigned int val = atoi( element->FirstChild()->Value() );
                if ( val == 0 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                if ( strstr( elementPath, "valueblock_1" ) )
                {
                    // Initialisierungswert angegeben
                    modifyData.hitagInitValue1 = val;
                    modifyData.flagsTranponder |= TMDF_INIT_VALUE1;
                }
                else if ( strstr( elementPath, "valueblock_2" ) )
                {
                    // Initialisierungswert angegeben
                    modifyData.hitagInitValue2 = val;
                    modifyData.flagsTranponder |= TMDF_INIT_VALUE2;
                }
                else
                {
                    // Initialisierungswert angegeben
                    modifyData.hitagInitValue3 = val;
                    modifyData.flagsTranponder |= TMDF_INIT_VALUE3;
                }
                break;
            }
            case TRANSPONDER_MIFARE: // Mifare 
            {
                std::string val = element->FirstChild()->Value();
                if ( strstr( elementPath, "valueblock_1" ) == NULL || strlen( val.c_str() ) > sizeof(modifyData.mifareInitalValue) )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }


                // Initialisierungswert angegeben
                strcpy( modifyData.mifareInitalValue, val.c_str() );
                modifyData.flagsTranponder |= TMDF_INIT_VALUE1;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/valueblock_1/autoinc" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_2/autoinc" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/valueblock_3/autoinc" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_HITAG2: // Hitag2 
            {

                // Flagvariable initialisieren
                if ( !(modifyData.flagsTranponder & TMDF_AUTOINC_FLAG1) && 
                    !(modifyData.flagsTranponder & TMDF_AUTOINC_FLAG2) &&
                    !(modifyData.flagsTranponder & TMDF_AUTOINC_FLAG3) )
                {
                    modifyData.hitagAutoincFlags = 0;
                }

                unsigned char flag = atoi( element->FirstChild()->Value() );
                if ( flag != 0 && flag != 1 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                if ( strstr( elementPath, "valueblock_1" ) )
                {
                    // Autoinc angegeben
                    modifyData.hitagAutoincFlags |= 0x01;
                    modifyData.flagsTranponder |= TMDF_AUTOINC_FLAG1;
                }
                else if ( strstr( elementPath, "valueblock_2" ) )
                {
                    // Autoinc angegeben
                    modifyData.hitagAutoincFlags |= 0x02;
                    modifyData.flagsTranponder |= TMDF_AUTOINC_FLAG2;
                }
                else
                {
                    // Autoinc angegeben
                    modifyData.hitagAutoincFlags |= 0x04;
                    modifyData.flagsTranponder |= TMDF_AUTOINC_FLAG3;
                }
                break;
            }
            case TRANSPONDER_MIFARE: // Mifare 
            {
                // Flagvariable initialisieren
                if ( !(modifyData.flagsTranponder & TMDF_AUTOINC_FLAG1) && 
                    !(modifyData.flagsTranponder & TMDF_AUTOINC_FLAG2) &&
                    !(modifyData.flagsTranponder & TMDF_AUTOINC_FLAG3) )
                {
                    modifyData.hitagAutoincFlags = 0;
                }

                unsigned char flag = atoi( element->FirstChild()->Value() );
                if ( strstr( elementPath, "valuevalueblock_1" ) != NULL && flag != 0 && flag != 1 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                // Autoinc angegeben
                modifyData.hitagAutoincFlags |= 0x01;
                modifyData.flagsTranponder |= TMDF_AUTOINC_FLAG1;
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/passwordgroup_1/modus" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_2/modus" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_3/modus" ) == 0 )
    {

        switch ( modifyData.technic )
        {
            case TRANSPONDER_MIFARE: // Mifare 
            {

                unsigned char val = atoi( element->FirstChild()->Value() );
                if ( val > 2 )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                if ( strstr( elementPath, "passwordgroup_1" ) )
                {
                    // Passwortmodus angegeben
                    modifyData.passwordFlags |= (val & 0x03);
                    modifyData.flagsPassword |= PMDF_MODUS1;
                }
                else if ( strstr( elementPath, "passwordgroup_2" ) )
                {
                    // Passwortmodus angegeben
                    modifyData.passwordFlags |= (val & 0x03) << 2;
                    modifyData.flagsPassword |= PMDF_MODUS2;
                }
                else
                {
                    // Passwortmodus angegeben
                    modifyData.passwordFlags |= (val & 0x03) << 4;
                    modifyData.flagsPassword |= PMDF_MODUS3;
                }
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/passwordgroup_1/key_a" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_2/key_a" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_3/key_a" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_1/key_b" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_2/key_b" ) == 0 ||
        strcmp( elementPath, "/modify/transponder/passwordgroup_3/key_b" ) == 0 )
    {
        switch ( modifyData.technic )
        {
            case TRANSPONDER_MIFARE: // Mifare
            {
                size_t i;
                std::string pass = element->FirstChild()->Value();
                for ( i = 0; i < strlen( pass.c_str() ); ++i )
                {
                    if ( !isxdigit( pass[i] ) )
                        break;
                }
                if ( !strlen( pass.c_str() ) || (i < strlen( pass.c_str() )) )
                {
                    // Undefiniertes Segment
                    level--;
                    return IV_ERROR_FILE_CONTENT_XML;
                }

                if ( strstr( elementPath, "passwordgroup_1/key_a" ) )
                {
                    // Passwort angegeben
                    strcpy( modifyData.mifarePasswordA0, pass.c_str() );
                    modifyData.flagsPassword |= PMDF_KEYA0;
                }
                else if ( strstr( elementPath, "passwordgroup_1/key_b" ) )
                {
                    // Passwort angegeben
                    strcpy( modifyData.mifarePasswordB0, pass.c_str() );
                    modifyData.flagsPassword |= PMDF_KEYB0;
                }
                else if ( strstr( elementPath, "passwordgroup_2/key_a" ) )
                {
                    // Passwort angegeben
                    strcpy( modifyData.mifarePasswordA1, pass.c_str() );
                    modifyData.flagsPassword |= PMDF_KEYA1;
                }
                else if ( strstr( elementPath, "passwordgroup_2/key_b" ) )
                {
                    // Passwort angegeben
                    strcpy( modifyData.mifarePasswordB1, pass.c_str() );
                    modifyData.flagsPassword |= PMDF_KEYB1;
                }
                else if ( strstr( elementPath, "passwordgroup_3/key_a" ) )
                {
                    // Passwort angegeben
                    strcpy( modifyData.mifarePasswordA2, pass.c_str() );
                    modifyData.flagsPassword |= PMDF_KEYA2;
                }
                else if ( strstr( elementPath, "passwordgroup_3/key_b" ) )
                {
                    // Passwort angegeben
                    strcpy( modifyData.mifarePasswordB2, pass.c_str() );
                    modifyData.flagsPassword |= PMDF_KEYB2;
                }
                break;
            }
            default:
            {
                // Unmoeglicher Wert fuer diesen Tranponder
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }
        }
    }
    else if ( strcmp( elementPath, "/modify/transponder/extern" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/transponder/extern/type" ) == 0 )
    {

        modifyData.extReader = atoi( element->FirstChild()->Value() );
        if ( modifyData.extReader != 0 && modifyData.extReader != 2 )
        {
            // Undefinierter Lesertyp
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
        modifyData.flags |= SMDF_TRANSPONDER_EXT_READER;

    }
    else if ( strcmp( elementPath, "/modify/password" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/password/system" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/password/system/value" ) == 0 )
    {   
        // Systempasswort uebernehmen
        modifyData.passwordSystem = atol( element->FirstChild()->Value() );
        if ( modifyData.passwordSystem > 9999 )
        {
            // Ungueltiger Wert 
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }

        // Passwort uebernommen
        modifyData.flags |= SMDF_PASSWORD_SYSTEM;

    }
    else if ( strcmp( elementPath, "/modify/password/user" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/password/user/value" ) == 0 )
    {   
        // Benutzerpasswort uebernehmen
        modifyData.passwordUser = atol( element->FirstChild()->Value() );
        if ( modifyData.passwordUser > 9999 )
        {
            // Ungueltiger Wert 
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }

        // Passwort uebernommen
        modifyData.flags |= SMDF_PASSWORD_USER;

    }
    else if ( strcmp( elementPath, "/modify/password/menu_item" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/password/menu_item/name" ) == 0 )
    {
        if ( password.name.size() == 0 )
        {
            password.name = element->FirstChild()->Value();
            if ( password.password != -1 )
            {
                modifyData.passwordMenu.push_back( password );
                modifyData.flags |= SMDF_PASSWORD_MENU;
                // Zuruecksetzen
                password.name.erase();
                password.password = -1;
            }
        }
        else
        {
            // Password ohne Wert
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }

    }
    else if ( strcmp( elementPath, "/modify/password/menu_item/value" ) == 0 )
    {    
        if ( password.password == -1 )
        {   
            password.password = atoi( element->FirstChild()->Value() );
            if ( password.password < 0 || password.password > 9999 )
            {
                // Ungueltiger Wert 
                level--;
                return IV_ERROR_FILE_CONTENT_XML;
            }

            if ( password.name.size() != 0 )
            {   
                modifyData.passwordMenu.push_back( password );
                modifyData.flags |= SMDF_PASSWORD_MENU;
                // Zuruecksetzen
                password.name.erase();
                password.password = -1;
            }
        }
        else
        {    
            // Password ohne Name
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
    }
    else if ( strcmp( elementPath, "/modify/fingerprint" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/fingerprint/security_level" ) == 0 )
    {
        modifyData.securityLevel = atoi( element->FirstChild()->Value() );
        if ( modifyData.securityLevel < 20 || modifyData.securityLevel > 79 ) // 20 - 79, default 60
        {
            // Auserhalb von Gueltigkeitsbereich
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
        modifyData.securityFRR = g_securityLevels[modifyData.securityLevel - 20][0];
        modifyData.securityFAR = g_securityLevels[modifyData.securityLevel - 20][1];
        modifyData.flags |= SMDF_FINGERPRINT_SECURITY_LEVEL;
    }
    else if ( strcmp( elementPath, "/modify/fingerprint/image_quality" ) == 0 )
    {    
        modifyData.quality = atoi( element->FirstChild()->Value() );
        if ( modifyData.quality < 10 || modifyData.quality > 100 || modifyData.quality % 5 ) // 10 - 100 in 5er Schritten, default 70
        {
            // Auserhalb von Gueltigkeitsbereich
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
        modifyData.flags |= SMDF_FINGERPRINT_QUALITY;
    }
    else if ( strcmp( elementPath, "/modify/fingerprint/minutiae_count" ) == 0 )
    {    
        modifyData.minutiaeCount = atoi( element->FirstChild()->Value() );
        if ( modifyData.minutiaeCount < 7 || modifyData.minutiaeCount > 25 ) // 7 - 25, default 20 
        {
            // Auserhalb von Gueltigkeitsbereich
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
        modifyData.flags |= SMDF_FINGERPRINT_MINUTIAE_COUNT;
    }
    else if ( strcmp( elementPath, "/modify/replace" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/replace/name" ) == 0 )
    {

    }
    else if ( strcmp( elementPath, "/modify/replace/name/search" ) == 0 )
    {
        if ( replaceString.search.size() == 0 )
        {
            replaceString.search = element->FirstChild()->Value();
            if ( replaceString.replacement.size() )
            {
                modifyData.replaceString.push_back( replaceString );
                modifyData.flags |= SMDF_REPLACE_STRING;
                replaceString.search.erase();
                replaceString.replacement.erase();
            }
        }
        else
        {
            // Replace ohne replacement
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
    }
    else if ( strcmp( elementPath, "/modify/replace/name/replacement" ) == 0 )
    {    
        if ( replaceString.replacement.size() == 0 )
        {
            replaceString.replacement = element->FirstChild()->Value();
            if ( replaceString.search.size() )
            {
                modifyData.replaceString.push_back( replaceString );
                modifyData.flags |= SMDF_REPLACE_STRING;
                replaceString.search.erase();
                replaceString.replacement.erase();
            }
        }
        else
        {
            // Replace ohne replacement
            level--;
            return IV_ERROR_FILE_CONTENT_XML;
        }
    }
    else
    {
        // Element unbekannt
        level--;
        return IV_ERROR_FILE_UNKNOWN_XML;
    }

    // Ende: Extrahieren der XML-Daten
    //###########################################

    int res;
    tinyxml2::XMLElement *node = element->FirstChildElement();
    while ( node != NULL )
    {
        if ( (res = Extract( elementPath, node, modifyData )) != IV_ERROR_SUCCESS )
        {    
            level--;
            return res;
        }

        node = node->NextSiblingElement();
    }

    level--;
    return IV_ERROR_SUCCESS;
}

